; //
; // ASMCrusher.asm - VST  
; //     (The trick), 2016
; //

format PE GUI 4.0 DLL at 11000000h

include 'win32wx.inc'

; //   VST 
struct AEEffect
    magic		    dd ?		; //  'VstP'
    dispatcher		    dd ?		; //  
    process		    dd ?
    setParameter	    dd ?		; //  
    getParameter	    dd ?		; //  
    numPrograms 	    dd ?
    numParams		    dd ?
    numInputs		    dd ?		; //   
    numOutputs		    dd ?		; //   
    flags		    dd ?		; // 
    resvd1		    dd ?
    resvd2		    dd ?
    initialDelay	    dd ?
    realQualities	    dd ?
    offQualities	    dd ?
    ioRatio		    dd ?
    object		    dd ?		; //    
    user		    dd ?
    uniqueId		    dd ?		; //   
    version		    dd ?		; //  
    processReplacing	    dd ?		; //   
    processDoubleReplacing  dd ?
    future		    db 56 dup (?)
ends

; //   ASMCrusher
struct ASMCrusher
    ae			    AEEffect ?		; //   AEEffect
    sampleRate		    dd ?		; //  
    volume		    dd ?		; //  0..1 (0..100%)
    downsampling	    dd ?		; //   0..1 (0..SampleRate)
    quantize		    dd ?		; //  0..1 (2 ^ (value * 15 + 1))
    lValue		    dd ?		; //     
    rValue		    dd ?		; //     
    sampleCounter	    dd ?		; //   
ends

NUMBER_OF_PARAMETERS	= 3			; //   
UNIQUE_ID		= 1234567		; //   
VERSION 		= 1			; //  
PAR_VOLUME		= 0			; //   ...
PAR_DOWNSAMPLING	= 1
PAR_QUANTIZE		= 2

kEffectMagic		= 0x56737450		; //  AEEffect
audioMasterVersion	= 1			; //  

; //   
kVstMaxParamStrLen	= 8
kVstMaxVendorStrLen	= 64
kVstMaxProductStrLen	= 64
kVstMaxEffectNameLen	= 32

effClose		= 1			; //     
effSetSampleRate	= 10			; //    
effGetParamName 	= 8			; //    
effGetParamLabel	= 6			; //      
effGetParamDisplay	= 7			; //     
effGetEffectName	= 45			; //    
effGetVendorString	= 47			; //    
effGetProductString	= 48			; //    
effGetVendorVersion	= 49			; //   

effFlagsCanReplacing	= 16
effFlagsNoSoundInStop	= 512

section '.idata' import data readable writeable

library kernel, 'kernel32.dll', \
	msvcrt, 'msvcrt.dll'

import kernel,\
       GetProcessHeap, 'GetProcessHeap', \
       HeapAlloc, 'HeapAlloc', \
       HeapFree, 'HeapFree', \
       lstrcpynA, 'lstrcpynA'

import msvcrt, \
       sprintf, 'sprintf'

data export
export 'AsmCrusher.DLL', Main, 'Main'
end data

section '.reloc' data readable discardable fixups
section '.text' code readable executable

EFFECT_NAME	 db	 'ASMCrusher', 0
VENDOR_NAME	 db	 '   (The trick)', 0
PRODUCT_NAME	 db	 'ASMCrusher', 0
PARAM_NAME_1	 db	 'Volume', 0
PARAM_NAME_2	 db	 'Frequency', 0
PARAM_NAME_3	 db	 'Quantize', 0
PARAM_LABEL_1	 db	 '%', 0
PARAM_LABEL_2	 db	 'Hz', 0
PARAM_LABEL_3	 db	 'Levels', 0
PARAM_FORMAT_1	 db	 '%d%%', 0
PARAM_FORMAT_2	 db	 '%dHz', 0
PARAM_FORMAT_3	 db	 '%d', 0

PARAMS_LIST	 dd	 PARAM_NAME_1, PARAM_NAME_2, PARAM_NAME_3	; //  
LABELS_LIST	 dd	 PARAM_LABEL_1, PARAM_LABEL_2, PARAM_LABEL_3	; //    
FORMATS_LIST	 dd	 PARAM_FORMAT_1, PARAM_FORMAT_2, PARAM_FORMAT_3 ; //  

entry EntryPoint

; //   DLL
proc EntryPoint, hinstDLL, fdwReason, lpvReserved
    mov eax, 1
    ret
endp

; //  
proc MemAlloc, size
    invoke HeapAlloc, <invoke GetProcessHeap>, HEAP_NO_SERIALIZE OR HEAP_ZERO_MEMORY, [size]
    ret
endp

; //  
proc MemFree, pMem
    invoke HeapFree, <invoke GetProcessHeap>, [pMem], HEAP_NO_SERIALIZE
    ret
endp

; //      VST 
proc Main c audioMaster

    ; //  
    cinvoke audioMaster, 0, audioMasterVersion, 0, 0, 0, 0
    .if eax = 0
	ret
    .endif

    stdcall CreateASMCrusher

    ret

endp

; //   ASMCrusher
proc CreateASMCrusher uses ebx

    stdcall MemAlloc, sizeof.ASMCrusher

    .if eax = 0
	ret
    .endif

    mov ebx, eax
    lea eax, [ebx + ASMCrusher.ae]

    mov [eax + ASMCrusher.ae.magic], kEffectMagic
    mov [eax + ASMCrusher.ae.dispatcher], Dispatcher
    mov [eax + ASMCrusher.ae.setParameter], SetParameter
    mov [eax + ASMCrusher.ae.getParameter], GetParameter
    mov [eax + ASMCrusher.ae.processReplacing], ProcessReplacing
    mov [eax + ASMCrusher.ae.numInputs], 2
    mov [eax + ASMCrusher.ae.numOutputs], 2
    mov [eax + ASMCrusher.ae.numParams], NUMBER_OF_PARAMETERS
    mov [eax + ASMCrusher.ae.flags], effFlagsCanReplacing OR effFlagsNoSoundInStop
    mov [eax + ASMCrusher.ae.uniqueId], UNIQUE_ID
    mov [eax + ASMCrusher.ae.version], VERSION
    mov [eax + ASMCrusher.ae.object], ebx

    ; //    
    mov [eax + ASMCrusher.sampleRate], 44100
    mov [eax + ASMCrusher.volume], 1.0
    mov [eax + ASMCrusher.downsampling], 1.0
    mov [eax + ASMCrusher.quantize], 1.0
    mov [eax + ASMCrusher.lValue], 0.0
    mov [eax + ASMCrusher.rValue], 0.0
    mov [eax + ASMCrusher.sampleCounter], 0.0

    mov eax, ebx

    ret

endp

; //  
proc Dispatcher c, pEffect, uOpcode, uIndex, value, lpPtr, opt

    mov ecx, [pEffect]
    mov ecx, [ecx + AEEffect.object]

    .if [uOpcode] = effClose
	; //  VST 
	stdcall MemFree, ecx
	xor eax, eax
    .elseif [uOpcode] = effSetSampleRate
	; //   
	mov eax, [opt]
	mov [ecx + ASMCrusher.sampleRate], eax
	xor eax, eax
    .elseif [uOpcode] = effGetParamName
	; //   
	mov eax, [uIndex]
	invoke lstrcpynA, [lpPtr], [PARAMS_LIST + eax * 4], kVstMaxParamStrLen
	xor eax, eax
    .elseif [uOpcode] = effGetParamLabel
	; //      ()
	mov eax, [uIndex]
	invoke lstrcpynA, [lpPtr], [PARAMS_LIST + eax * 4], kVstMaxParamStrLen
	xor eax, eax
    .elseif [uOpcode] = effGetEffectName
	; //   
	invoke lstrcpynA, [lpPtr], EFFECT_NAME, kVstMaxEffectNameLen
	xor eax, eax
    .elseif [uOpcode] = effGetVendorString
	; //   
	invoke lstrcpynA, [lpPtr], VENDOR_NAME, kVstMaxVendorStrLen
	xor eax, eax
    .elseif [uOpcode] = effGetProductString
	; //   
	invoke lstrcpynA, [lpPtr], PRODUCT_NAME, kVstMaxProductStrLen
	xor eax, eax
    .elseif [uOpcode] = effGetVendorVersion
	; //  
	mov eax, VERSION
    .elseif [uOpcode] = effGetParamDisplay
	; //    ()
	.if [uIndex] = PAR_VOLUME

	    ; // volume * 100
	    mov eax, 100.0
	    movd xmm0, eax
	    mulss xmm0, [ecx + ASMCrusher.volume]
	    cvtss2si eax, xmm0

	    cinvoke sprintf, [lpPtr], [FORMATS_LIST + PAR_VOLUME * 4], eax

	.elseif [uIndex] = PAR_DOWNSAMPLING

	    stdcall CalcDownsamplingFreq, [ecx + ASMCrusher.sampleRate], [ecx + ASMCrusher.downsampling]
	    cinvoke sprintf, [lpPtr], [FORMATS_LIST + PAR_DOWNSAMPLING * 4], eax

	.elseif [uIndex] = PAR_QUANTIZE

	    stdcall CalcLevels, [ecx + ASMCrusher.quantize]
	    cinvoke sprintf, [lpPtr], [FORMATS_LIST + PAR_QUANTIZE * 4], eax

	.endif

	xor eax, eax
    .else
	xor eax, eax
    .endif

    ret

endp

; //  
proc GetParameter c, pEffect, uIndex

    mov ecx, [pEffect]
    mov ecx, [ecx + AEEffect.object]

    .if [uIndex] = PAR_VOLUME
	fld [ecx + ASMCrusher.volume]
    .elseif [uIndex] = PAR_DOWNSAMPLING
	fld [ecx + ASMCrusher.downsampling]
    .elseif [uIndex] = PAR_QUANTIZE
	fld [ecx + ASMCrusher.quantize]
    .else
	fldz
    .endif

    ret
endp

; //  
proc SetParameter c, pEffect, uIndex, fValue

    mov ecx, [pEffect]
    mov ecx, [ecx + AEEffect.object]
    mov eax, dword [fValue]

    .if [uIndex] = PAR_VOLUME
	mov [ecx + ASMCrusher.volume], eax
    .elseif [uIndex] = PAR_DOWNSAMPLING
	mov [ecx + ASMCrusher.downsampling], eax
    .elseif [uIndex] = PAR_QUANTIZE
	mov [ecx + ASMCrusher.quantize], eax
    .endif

    ret

endp

; //   ,  
proc ProcessReplacing c uses esi edi ebx, pEffect, pInputs, pOutputs, sampleFrames

    mov esi, [pInputs]
    mov edi, [pOutputs]
    mov ebx, [pEffect]
    mov ebx, [ebx + AEEffect.object]

    ; //   
    stdcall ApplyEffectToChannel, ebx, dword [esi], dword [edi], dword [sampleFrames], dword [ebx + ASMCrusher.lValue]
    mov [ebx + ASMCrusher.lValue], eax

    ; //   
    stdcall ApplyEffectToChannel, ebx, dword [esi + 4], dword [edi + 4], dword [sampleFrames], dword [ebx + ASMCrusher.rValue]
    mov [ebx + ASMCrusher.rValue], eax

    ret

endp

; //    
; //   
proc ApplyEffectToChannel uses esi edi ebx, pObject, pInput, pOutput, nCount, fValue

    mov ebx, [pObject]
    mov esi, [pInput]
    mov edi, [pOutput]

    ; //   
    stdcall CalcLevels, [ebx + ASMCrusher.quantize]
    cvtsi2ss xmm1, eax

    ; //      
    mov eax, 2.0
    movd xmm0, eax
    divss xmm0, [ebx + ASMCrusher.downsampling]

    ; //    
    movss xmm2, [ebx + ASMCrusher.sampleCounter]

    ; //   
    mov eax, 1.0
    movd xmm3, eax
    mov eax, -1.0
    movd xmm4, eax

    ; //     
    movss xmm6, [ebx + ASMCrusher.volume]

    ; //        
    movd xmm5, [fValue]
    mulss xmm5, xmm6

    ; //   
    mov ecx, [nCount]

    ; //   
    .PROCESS_SAMLE:

	; //    
	addss xmm2, xmm3
	comiss xmm2, xmm0
	; //     ,  
	jb .STORE_SAMPLE

	; //   1
	comiss xmm3, dword [esi]
	jb .SET_MAX
	; //   -1
	comiss xmm4, dword [esi]
	ja .SET_MIN
	movss xmm5, dword [esi]

	.CALC_SAMPLE:
	; //     edx
	movd edx, xmm5
	; //     int(sample * levels) / levels
	mulss xmm5, xmm1
	cvtss2si eax, xmm5
	cvtsi2ss xmm5, eax
	divss xmm5, xmm1

	; //  
	mulss xmm5, xmm6
	; //  downsampling 
	subss xmm2, xmm0

	jmp .STORE_SAMPLE

	.SET_MAX:
	movss xmm5, xmm3
	jmp .CALC_SAMPLE
	.SET_MIN:
	movss xmm5, xmm4
	jmp .CALC_SAMPLE

	.STORE_SAMPLE:

	; //   
	movss dword [edi], xmm5
	add edi, 4
	add esi, 4

    loop .PROCESS_SAMLE

    ; //  
    movd [ebx + ASMCrusher.sampleCounter], xmm2

    ; //   
    mov eax, edx

    ret

endp

; //        downsampling
; //    int(downsampling * samplerate * 0.5)
proc CalcDownsamplingFreq, sampleRate, downsampling

    mov eax, 0.5
    movd xmm0, eax
    mulss xmm0, [downsampling]
    mulss xmm0, [sampleRate]
    cvtss2si eax, xmm0

    ret

endp

; //        quantize
; //    int(2 ^ (quantize * 15 + 1)))
proc CalcLevels, quantize

    mov eax, 2
    mov ecx, 15.0
    movss xmm0, [quantize]
    movd xmm1, ecx
    mulss xmm0, xmm1
    cvtss2si ecx, xmm0
    shl eax, cl

    ret

endp
